'==============================================================================
'Filename   Si570_LO_2.02b.bas
'Version    2.02b for use with PCB V2.00 (03/2011)
'Date       2014-10-12
'Purpose    Local oscillator for TRX with band control using Si570
'           Parts of Si570 code from Sid Knox, W7QJQ - THANKS!
'           upgraded to allow smooth tuning within +/- 3,500 ppm LO frequency
'Author     Georg Latzel, DL6GL
'Copyright  (c) Georg Latzel, free for private use under GNU license
'Hardware   ATmega 32, Si570B (LVDS), LCD 16x2, rotary encoder, 3 keys
'
'           ALPS STEC11B04 encoder (reichelt.de)
'           Turning clockwise              ... counter clockwise
'           A    B                             A     B
'           1    1  start detent position      1     1
'           1    0  intermediate position 1    0     1
'           0    0  intermediate position 2    0     0
'           0    1  intermediate position 3    1     0
'           1    1  next detent position       1     1
'           Set DoubleStep = 0 (see below)

'           ALPS EC11E15244B2 encoder (pollin.de some years ago):
'           Turning clockwise              ... counter clockwise
'           A    B                             A     B
'           1    1  start detent position      1     1
'           0    1  intermediate position 1    1     0
'           0    0  intermediate position 2    0     0
'           1    0  intermediate position 3    0     1
'           1    1  intermediate position 4    1     1
'           0    1  intermediate position 5    1     0
'           0    0  next detent position       0     0
'           Detent positions are different (1/1) or (0/0) resp.
'           Set DoubleStep = 1 (see below)

'           DoubleStep is set on the front panel board contact strip
'           DoubleStep = 0: Connect contact strip pin 20 to adjacent GND pin
'                           (solder bridge), is the former "Spare" pin
'           DoubleStep = 1: Nothing to do, was the standard configuration
'                           in firmware V2.00, V2.01, V3.00 to V3.04


'Compiler   BASCOM-AVR 2.0.7.6
'           For readable "camel case" variables mark "Don't change case"
'           in BASCOM Options-Programmer-Environment-Editor
'Edit history
'V1.00      2010-10-01, first release
'V1.01      2010-10-10  added: last frequency is stored in array lngLastFreq
'                       so last freq is remembered for each band
'V1.02      2010-11-01  added:
'                       LO freq is changed if TRX data are changed
'                       Freq. step for xtal calibration can be changed
'                       in the same way as freq. step for tuning
'V2.00      2011-04-08  Introduced: configuration file Si570_LO_2.00_Config.bas
'                       Setting of complete VFO data, see $Include
'                       Added: FM IF filter, center freq. can be different
'                       Changed: EEPROM data structure
'                       Version a:  PCB version 1 (10/2010)
'                                   FM switch could be added to PC2
'                                   SCL on PC3
'                                   SDA on PC4
'                       Version b:  PCB version 2 (03/2011)
'                                   SCL on PC2
'                                   SDA on PC3
'                                   FM switch on PC4
'                       Enable/disable corresponding lines, see comments
'                       Flash used: 68%
'V2.01      2011-12-22  HW interrupt INT1 for encoder does not work properly
'                       Now polling by 8 bit Timer0 overflow
'                       Flash used: 69%
'V2.02      2014-10-12  Faster Timer0 polling of keys & encoder (~ 500Hz)
'                       Setting of double step option on front panel board
'                       Updates V3.01, 3.02, 3.03 included
'                       V3.01:To fit ALPS STEC11B13 encoder (reichelt.de)
'                             evaluate intermediate positions or not
'                             Const DoubleStep = 0 / =1, see above
'                             Timer0_isr changed accordingly
'                             Stop Timer0 while displaying encoder data.
'                       V3.02:BPF & LPF setting just after I2C check
'                             PCF8574 has high output after power on
'                             so all relais are powered by ULN2803.
'                       V3.03:New key handling (Dannegger type with Timer0 interrupt)
'                             External key interrupt via INT0 cancelled
'                             Diodes from keys to INT0 (PD.2) need not be removed
'                       Flash used: 22,502 bytes (68%)
'==============================================================================

'Internal Timer0 interrupt is used for encoder & Keys

'2 main menu levels
'Menu level 1: normal operation, set by encoder key
'Menu level 2: settings, set by store key, pressed short
' (1.1)  set freq via encoder in RX mode (locked in TX mode)
'        set bands via up-/down keys in RX mode (locked in TX mode)
'        At the band limits "<" or ">" is shown if in
'        - SSB mode
'          TRX frequency - 1.5 kHz is below band limit (LSB)
'          TRX frequency + 1.5 kHz is above land limit (USB)
'        - FM mode
'          TRX frequency - 6.0 kHz is below band limit
'          TRX frequency + 6.0 kHz is above land limit
'        press encoder key to go to (1,2)
' (1.2)  set freq tuning steps via encoder
'        press encoder key to go to (1,1)
' (2  )  level 2 entered from level 1.x by pressing sto key long
' (2,1)  set SSB IF filter freq via encoder, set digits via up/down keys
'        press store key long to save setting
'        press store key short to go to (2,2)
' (2,2)  set LSB offset via encoder, set digits via up/down keys
'        press store key long to save setting
'        press store key short to go to (2,3)
' (2,3)  set USB offset via encoder, set digits via up/down keys
'        press store key long to save setting
'        press encoder key to go to (2,4)
' (2,4)  set CW IF filter freq, same as SSB IF filter
' (2,5)  set CW offset via encoder, set digits via up/down keys
'        press store key long to save setting
'        press store key short to go to (2,5)
' (2,6)  set FM IF filter freq, same as SSB IF filter
' (2,7)  set Mixer divider via encoder (1, 2, 4, 8)
'        Freq-LO = (Freq-RTX + Freq IF) * divider factor
'        press store key long to save setting
'        press store key short to go to (2,6)
' (2,8)  set Si570 slave address
'        press store key long to save setting
'        press store key short to go to (2,7)
' (2,9)  calibrate Si570 xtal frequency (114.285 MHz +/- 2,000 ppm)
'        could be done stand alone before connecting to the mixer
'        - LO frequency ist set to 10 MHz (WWV)
'        - connect a short wire to LO output
'        - listen to the beat tone in your RX tuned to WWV
'        or use a calibrated frequency counter,
'        set xtal freq. via encoder up/down to get 10 MHz at the LO output,
'        Increasing xtal freq. results in decreasing LO freq.
'        To change freq step press encoder button similar to menu 1.2.
'        Press store key long to save setting.
'        Press store key short to go to (3.0).
'        Check calibration with the LO connected to the mixer:
'        For xtal filter center freq. 9 MHz and LSB/USB offsets -/+ 1.5 kHz
'        LSB: tune to 10,001,500, WWV beat tone should be ~ 1.5 kHz
'             LO freq.= 19,000,000 MHz
'        USB: tune to  9,998,500, WWV beat tone should be ~ 1.5 kHz
'             LO freq.= 19,000,000 MHz
' (3,0) Set bandpass filter data
'        Row 1: BPF used = 1 / not used = 0
'        Row 2: BPF sequential number
'        Toggle row1/2 with encoder key
'        Press store key long to save setting.
'        Press store key short to go to (1.1)

'Bandpass filters are selected via 2 PCF8574A (type A!)
'Bank 1  Amateur bands 1...8, 160/80/60/40/30/20/17/15m
'        PCF8574A address 0111.000.0 = 70 hex
'        0111 = factory setting for PCD8574A
'        000  = A0=A1=A2=0 (gnd)
'Bank 2  further amateur bands, 12/10/6/2 m
'        PCF8574A address 0111.001.0 = 72 hex
'        001  = A0=1 (+5V), A1=A2=0 (gnd)
'Use of bank1 and bank2 is configured according to the number of used bands
'i.e. for merely 8 bands only bank 1 (1 PCF8574A) is used.
'2 adjacent bands can share 1 bandpass filter, e.g. 17+15m, 12+10m
'See comments in ...config.bas file.

'Settings and band data are stored in EEPROM
'For a new chip EEPROM is initialized with data from ...config.bas file, e.g.:
'lngStep       1.0 kHz        Freq. tuning step
'lngIFSSB      9 MHz          IF center freq. SSB
'lngIFCW       9 MHz          IF center freq. CW
'lngIFFM       9 MHz          IF center freq. FM
'wrdLSBOff     1.5 kHz        LSB offset relative to lngIF
'bytLSBSign    1              Subtract offset
'wrdUSBOff     1.5 kHz        LSB offset relative to lngIF
'bytUSBSign    2              add offset
'wrdCWOff      0.8 kHz        CW offset relative to lngIF
'bytCWSign     2              add offset
'bytLODiv      1              mixer divider factor
'bytAddrSi570  hex55          Si570 factory setting slave address
'dblFxtal      114.320760 MHz internal xtal freq. for used Si570
'bytI2C_BPF1   B01110000      Write Addr. PCF8574A bandpass filter bank 1
'bytI2C_BPF2   B01110010      Write Addr. PCF8574A bandpass filter bank 2
'EEPROM can be re-initialized by pressing store key long on power up

'FUSE-Bits (Ponyprog-Settings, [ ]=no check, [x]=check)
'In ATmega data sheets: Bit = 1 -> not set = [ ] Ponyprog
'                       Bit = 0 -> set     = [x] Ponyprog
'- OCDEN    [ ]
'- JTAGEN:  [ ] (disabled!, otherwise LCD at PORT.C does not work)
'- EESAVE   [ ] Save EEProm on erase chip
'- BOOTSZ1  [x] Select Boot size
'- BOOTSZ0  [x] Select Boot size
'- BOOTRST  [ ] Select reset vector
'- Crystal oscillator
'  CKOPT:   [ ] (=1, low Oscillator amplitude or=1, [x], high ampl.)
'  CKSEL3:  [ ] (=1, Select Clock Source)
'  CKSEL2:  [ ] (=1, as well)
'  CKSEL1:  [ ] (=1, as well)
'  CKSEL0:  [ ] (=1, crystal osc. fast start up 16k CK + 4ms)
'  SUT1:    [ ] (=1, as well)
'  SUT0:    [x] (=0, as well)

$regfile = "m32def.dat"                           'ATmega32
$crystal = 16000000                               '16 MHz
$hwstack = 80
$swstack = 80
$framesize = 100
Const txtVers = "Si570 LO V2.02b"                 'Version number

'Configure Ports - Port A (not used) ------------------------------------------
DDRA = &B00000000                                 'PortA = input
PORTA = &B11111111                                'Activate Pullup

'Configure LCD on Port B ------------------------------------------------------
Config Lcd = 16 * 2
Config Lcdpin = Pin , Db7 = PortB.4 , Db6 = PortB.3 , Db5 = PortB.2
Config Lcdpin = Pin , Db4 = PortB.5 , E = PortB.1 , Rs = PortB.0
Config Lcdbus = 4

'Configure Ports - Port C (RX/TX, SSB/CW/FM, LSB/USB) -------------------------
'Version a:  PCB version 1 (10/2010) ******************************************
'            FM switch could be added to PC2
'            See I2C configuration below
'            Enable the next 3 lines
'DDRC.2 = 0                                        'Input, TRX data (FM)
'PORTC.2 = 1                                       'Activate Pullup
'bitFM Alias PINC.2                                'PINC.2, SSB or CW=Hi, FM=Lo

'Version b:  PCB version 2 (03/2011) ******************************************
'            FM switch on PC4
'            See I2C configuration below
'            Enable the next 3 lines
DDRC.4 = 0                                        'Input, TRX data (FM)
PORTC.4 = 1                                       'Activate Pullup
bitFM Alias PINC.4                                'PINC.4, SSB or CW=Hi, FM=Lo

DDRC.7 = 0                                        'Input, TRX data (LSB/USB)
DDRC.6 = 0                                        'Input, TRX data (SSB/CW)
DDRC.5 = 0                                        'Input, TRX data (RX/TX)
PORTC.7 = 1                                       'Activate Pullup
PORTC.6 = 1
PORTC.5 = 1
bitLSB Alias PINC.7                               'PINC.7, LSB=Hi, USB=Lo
bitSSB Alias PINC.6                               'PINC.6, SSB=Hi, CW=Lo
bitPTT Alias PINC.5                               'PINC.5=PTT, RX: Hi, TX: Lo

'Configure I2C on Port C ------------------------------------------------------
'Bascom I2C/TWI software emulation is used, ports can be chosen arbitrarily
'Version a:  PCB version 1 (10/2010) ******************************************
'            SCL on PC3
'            SDA on PC4
'            Enable the next 2 lines
'Config SCL = PortC.3                              'I2C, SCL
'Config SDA = PortC.4                              'I2C, SDA

'Version b:  PCB version 2 (03/2011) ******************************************
'            SCL on PC2
'            SDA on PC3
'            Enable the next 2 lines
Config SCL = PortC.2                              'I2C, SCL
Config SDA = PortC.3                              'I2C, SDA

'Configure Encoder & keys - Portc C & D ---------------------------------------
DDRC.1 = 0                                        'PINC.1 = double step option
DDRC.0 = 0                                        'PINC.0 = Key store
PORTC.1 = 1                                       'Activate Pullup
PORTC.0 = 1
DDRD = &B00000000                                 'PortD = input
PORTD = &B11111111                                'Activate Pullup
'                                                 'PIND.7 = Key up
'                                                 'PIND.6 = Key down
'                                                 'PIND.5 = Encoder key
'                                                 'PIND.4 = Encoder Pin B
'                                                 'PIND.3 = Encoder Pin A (INT1)
'                                                 'PIND.2 = All keys (INT0)
'Port assignments for Timer0_isr:
Dim Key_port As Byte
Const Key_sto = 0                                 'Key store @ PINC.0
Const Key_up = 1                                  'Key up    @ PIND.7
Const Key_dwn = 2                                 'Key dwn   @ PIND.6
Const Key_enc = 3                                 'Enc Key   @ PIND.5
Enc_B Alias PIND.4                                'Encoder B
Enc_A Alias PIND.3                                'Encoder A
Pin_KeySto Alias PINC.0                           'Key store
Pin_DoubleStep Alias PINC.1                       'Encoder double step
Pin_KeyUp Alias PIND.7                            'Key up
Pin_Keydwn Alias PIND.6                           'Key down
Pin_KeyEnc Alias PIND.5                           'Encoder key

'Configure Timer0-Interrupt for Keys (Timer0 = 8bit, 2^8=256 counts) ----------
Config Timer0 = Timer , Prescale = 256
Const Presettimer0 = 130                          'Reduce timer0: 256-130=126 counts
'Overflow time: Overflow-Counts * Prescale / crystal frequency
'Possible Prescales: 8, 64, 256, 1024, here 1024
'= 126 * 256 / 16.000.000 = 2.02 msec = ~ 500 Hz (crystal 16 MHz)
On Timer0 Timer0_isr                              'On Overflow go to Timer0_isr
Const wrdTimer0_long = 500                        'Timer counts for long press 1 sec

'Variables Program control ----------------------------------------------------
Dim lngLO As Long                                 'LO frequency (Hz)
Dim lngFreq As Long                               'TRX frequency (Hz)
Dim lngStep As Long                               'Frequency tuning step (Hz)
Dim lngFreqSave As Long
Dim lngIFFreq As Long                             'SSB/CW/FM filter center freq.
Dim lngLastFreq(10) As Long                       'last freq. per band 1...10
Dim bytCursorPos As Byte                          'LCD Cursor position freq. setting
Dim bytCursorMin As Byte
Dim bytCursorMax As Byte
Dim bytSignPos As Byte                            'LCD position of kHz sign
Dim bytSign As Byte                               'Sign (1=-/2=+) of offset
Dim bytMenu As Byte                               'Menu number
Dim bytMenuBPF As Byte                            'Last menu number for BPF
Dim bytNumChar As Byte                            '# of characters frequency
Dim bytBPFNum As Byte                             '# of current bandpass filter
Dim bitStepSet As Bit                             '=1: tuning step was changed
Dim bitLSBOld As Bit                              'LSB/USB setting before change
Dim bitSSBOld As Bit                              'SSB/CW setting before change
Dim bitPTTOld As Bit                              'RX/TX setting before change
Dim bitFMOld As Bit                               'SSB or CW/FM setting before change
Dim bitTRXDataChanged As Bit                      '=1: TRX data were changed

'Variables EEPROM settings ----------------------------------------------------
Dim lngIFSSB As Long                              'SSB IF filter center frequency (Hz)
Dim lngIFCW As Long                               'CW IF filter center frequency (Hz)
Dim lngIFFM As Long                               'FM IF filter frequency (Hz)
Dim wrdLSBOff As Word                             'Offset LSB relative to lngIF (Hz)
Dim wrdUSBOff As Word                             'Offset USB relative to lngIF (Hz)
Dim bytLSBSign As Byte                            'Sign (1=-/2=+) LSB offset
Dim bytUSBSign As Byte                            'Sign (1=-/2=+) USB offset
Dim wrdCWOff As Word                              'Offset CW relative to lngIF (Hz)
Dim bytCWSign As Byte                             'Sign (1=-/2=+) CW offset
Dim bytLODiv As Byte                              'LO divide factor, 1, 2, 4, 8
Dim bytI2C_Adr As Byte                            'I2C Address
Dim wrdAddr As Word                               'EEPROM address
Dim bytPart(4) As Byte                            '4 bytes for a long variable


'Variables keys ---------------------------------------------------------------
Dim bytIwr0 As Byte
Dim bytKey_ct0 As Byte
Dim bytKey_ct1 As Byte
Dim bytKey_state As Byte
Dim bytKey_press As Byte
Dim bytKey_save As Byte
Dim wrdTimer0_runs As Word
Dim bytKey_mask As Byte
bytKey_Mask = &B00000001                          'Mask for short/long keys
'                                                  Set 1 to enable short/long

Dim bitKey_up As Bit                              '=1 if key up is pressed
Dim bitKey_dwn As Bit                             '=1 if key down is pressed
Dim bitKey_enc As Bit                             '=1 if key encoder is pressed
Dim bytKey_sto As Byte                            'Sto key pressed: 1(short), 2(long)
Const Key_Short = 1                               'Sto key pressed short
Const Key_Long = 2                                'Sto key pressed long

'Variables Encoder  -----------------------------------------------------------
Dim bytIncr As Byte                               '=2: Encoder incr., =1: decr., =0: ./.
Dim bytEnc_old As Byte
Dim bytEnc_new As Byte

'Common Variables, multiply used ----------------------------------------------
Dim bitTmp0 As Bit
Dim bytTmp0 As Byte
Dim bytTmp1 As Byte
Dim bytTmp2 As Byte
Dim bytSave As Byte
Dim intTmp0 As Integer
Dim intTmp1 As Integer
Dim intTmp2 As Integer
Dim intTmp3 As Integer
Dim wrdTmp0 As Word
Dim wrdTmp1 As Word
Dim wrdSave As Word
Dim lngTmp0 As Long
Dim lngTmp1 As Long
Dim lngSave As Long
Dim sngTmp0 As Single
Dim dblTmp0 As Double
Dim dblSave As Double
Dim str1 As String * 1
Dim str2 As String * 2
Dim str4 As String * 4
Dim str10 As String * 10
Dim str11 As String * 11

'Variables Bands & bandpass filter --------------------------------------------
Dim bitBPFUsed As Bit
Dim bytBPFCount As Byte
Dim bytI2C_BPF1 As Byte                           'I2C write addr. PCF8574A BPF 1
Dim bytI2C_BPF2 As Byte                           'I2C write addr. PCF8574A BPF 2
Dim bytBPFStatus As Byte                          'Status (0/1) for BPF relais
Dim bytActBand As Byte                            'Active band index
Dim bytBands As Byte                              'Number of used bands (<= 12)
Dim bytBPFs As Byte                               'Number of used bandpass filters
Dim bytBandUsed(12) As Byte                       '=1: band used, =0: not used
Dim bytBand(12) As Byte                           'band wavelength (meters)
Dim bytBandBPF(12) As Byte                        'Number of BPF for this band
Dim lngBandLo(12) As Long                         'Low  band limits 160-2m (12 bands)
Dim lngBandHi(12) As Long                         'High band limits 160-2m (12 bands)
Const bytBandsMax = 12                            'Bands arrays length

'Variables Si570 --------------------------------------------------------------
Const lngFNom = 5260                              'nominal mid-point Si570 osc
Const dbl2pwr28 = 268435456                       '2^28
Dim bitSmooth As Bit                              '=1: smooth tuning,
'                                                   =0: complete tuning calculation
Dim bytAddrSi570 As Byte                          'Si570 slave address
Dim bytSi570WriteAdr As Byte                      'Si570 I2C write mode address
Dim bytSi570ReadAdr As Byte                       'Si570 I2C read  mode address
Dim bytFinalHS_DIV As Byte                        'final value HS_DIV
Dim bytN1 As Byte
Dim bytHS_DIV(6) As Byte                          'array of valid HS_DIV values
Dim bytWriteRegs(6) As Byte                       'Si570 write register array
Dim bytArrayIndex As Byte                         'array pointer

Dim lngFout As Long                               'LO freq (Hz)
Dim lngConstant As Long
Dim lngCheckFreq As Long

Dim lngCenterFreq As Long                         'Current center freq configuration
Dim lngLimitHi As Long                            'Upper limit smooth tuning
Dim lngLimitLo As Long                            'Lower limit smooth tuning
Dim lngFxtal As Long                              'Si570 internal xtal freq (Hz)
Dim dblFxtal As Double                            'Si570 internal xtal freq (MHz)
Dim dblFout As Double                             'output freq, MHz
Dim dblFoutOld As Double                          'output freq, MHz
Dim dblRFREQ As Double                            'DCO freq control
Dim dbIintRFREQ As Double                         'DCO freq control,integer
Dim dblFracRFREQ As Double                        'DCO freq control,fraction
Dim dblHS_DIVXN1 As Double                        'DCO divisor constant

'Variables Si570 calibration  -------------------------------------------------
Dim bitCal As Bit                                 '=0: set xtal freq, =1: set freq step
Dim lngStepCal As Long                            'Frequency tuning step (Hz)

'Include configuration file ---------------------------------------------------
$Include "Si570_LO_2.00_Config.bas"

'Subroutines & Functions ------------------------------------------------------
'Operation
Declare Function KeyPressedLong(byVal bytKey_Long As Byte) As Byte
Declare Sub ClearLCDRow(byVal bytRow As Byte)
Declare Sub ShowMHz(ByVal lngF As Long , byVal bytRow As Byte)
Declare Sub ShowkHz(ByVal lngF As Long , byVal bytRow As Byte)
Declare Sub PrepareIFOffset(byVal wrdOffFreq As Word , byVal bytOffSign As Byte)
Declare Sub PrepareIFCenter(byVal lngIFFreq As Long)
Declare Function CheckI2C(byVal bytAdr As Byte) As Byte
Declare Sub WriteI2C(byVal bytAdr As Byte , bytByte As Byte)
Declare Sub SetBPFRelais(byVal bytBandNum As Byte , byVal bytStat As Byte)
Declare Function CalcLOFreq(ByVal lngTRX As Long) As Long
Declare Sub TRXData
Declare Sub CheckBandLimits(byVal lngF As Long)
Declare Sub SetGetLastFreq(byVal bytOpt As Byte)
'Settings
Declare Function GetCursorPos(lngF As Long) As Byte
Declare Function SetIFfreq(byVal lngF As Long , byVal bytDigit As Byte) As Long
Declare Function SetValue(byVal lngVal As Long) As Long
'EEPROM
Declare Sub EEPROMWriteLng(lngLong As Long)
Declare Sub EEPROMReadLng(lngLong As Long)
Declare Sub EEPROMWriteWrd(wrdWord As Word)
Declare Sub EEPROMReadWrd(wrdWord As Word)
Declare Sub EEPROMInit
Declare Sub EEPROMReadSettings
Declare Sub EEPROMWriteSettings
Declare Sub EEPROMReadBands(byVal bytAll As Byte)
'Si570
Declare Sub CalcSi570Addr
Declare Sub CalcConstant
Declare Sub CalcRFREQ
Declare Sub PackRegData
Declare Sub WriteSi570regs
Declare Sub SetNewFrequency
Declare Sub SetSmoothLimits
Declare Sub CheckSmoothLimits

Cls
'Init EEPROM data for a fresh flashed AVR -------------------------------------
'Press store key at power up to force a new initialization
'Read flashed EEPROM data in any case
'On power up all BPF PCF8574 outputs are high, so set BPF relais
Call EEPROMInit                                   'Initialize EEPROM if empty
Call EEPROMReadSettings                           'Read settings from EEPROM
Call EEPROMReadBands(0)                           'Read band data from EEPROM
Call SetBPFRelais(1 , 1)                          'Set BPF relais via I2C
Call CalcSi570Addr                                'Read/write adresses Si570

Cls
Cursor Off
Locate 1 , 1
Lcd txtVers

'Check I2C units (Si570 & bandpass filter) ------------------------------------
bytTmp2 = 0
bytTmp1 = 3                                       'Si570 & 2 PCF8574 (> 8 BPF)
If bytBPFs < 9 Then                               'Si570 & 1 PCF8574 (<=8 BPF)
   bytTmp1 = 2
End If

For bytTmp0 = 1 To bytTmp1
   Select Case bytTmp0
      Case 1                                      'Check Si570
         bytI2C_Adr = bytSi570WriteAdr
         str10 = "I2C Si570 "
      Case 2                                      'Check BPF bank 1
         bytI2C_Adr = bytI2C_BPF1
         str10 = "I2C BPF1  "
      Case 3                                      'Check BPF bank 2
         bytI2C_Adr = bytI2C_BPF2
         str10 = "I2C BPF2  "
   End Select
   Select Case CheckI2C(bytI2C_Adr)               'Check I2C @ bytI2C_Adr
      Case 0
         Locate 2 , 1
         LCD str10 ; "NOT OK"
         Wait 1
         bytTmp2 = 1
   End Select
Next bytTmp0
If bytTmp2 = 1 Then
   Wait 1
End If
CLS

'Initialize Si570 data --------------------------------------------------------
bytHS_DIV(1) = 11                                 'valid Si570 HS_DIV values
bytHS_DIV(2) = 9                                  'in decending order
bytHS_DIV(3) = 7                                  'DO NOT CHANGE!
bytHS_DIV(4) = 6                                  'DO NOT CHANGE!
bytHS_DIV(5) = 5                                  'DO NOT CHANGE!
bytHS_DIV(6) = 4                                  'DO NOT CHANGE!

'Initialize Key Data ----------------------------------------------------------
bytEnc_old = 0
bytEnc_new = 0
bytKey_ct0 = 0
bytKey_ct1 = 0
Decr bytKey_ct0
bytKey_ct1 = bytKey_ct0
bytKey_state = 0
bytKey_press = 0
bytKey_sto = 0

'Initialize LO for the first band ---------------------------------------------
bytActBand = 1                                    'First band (160m)
bitCal = 0                                        'normal xtal calibration mode
lngFreq = lngBandLo(bytActBand)                   'Band start freq
lngLO = CalcLOFreq(lngFreq)                       'Start LO freq.
For bytTmp0 = 1 To bytBands                       'Initialize array lngLastFreq
   lngLastFreq(bytTmp0) = 0
Next bytTmp0
Call SetNewFrequency                              'New frequency to Si570
Call SetSmoothLimits                              'Set smooth tuning limits
Call ShowMHz(lngFreq , 2)                         'show frequency on LCD
Call CheckBandLimits(lngFreq)

'Start of Main program ********************************************************

Enable Timer0                                     'Interrupt encoder
Enable Interrupts                                 'enable all Interrupts
Timer0 = Presettimer0                             'Initialize Timer0

lngStepCal = 1000                                 '1 kHz step for Si570 calibration
bytMenu = 11                                      'Start menu: Show frequency

Do

   'Check keys, any key pressed? ----------------------------------------------
   'bytKey_press returned from Timer0_isr, flag bits 0...7 = keys 0...7
   'Just pressed detected in Timer0_isr: pressed status flag bit = 1
   'After evaluating key pressed status flag, bit must be reset
   'Short/long pressed status flag is reset in function KeyPressedLong
   'Key action is stored in the variables:
   '-  bitKey_up     =1: up key was pressed,      =0: else
   '-  bitKey_dwn    =1: down key was pressed,    =0: else
   '-  bitKey_enc    =1: encoder key was pressed, =0: else
   '-  bytKey_sto    =0: store key was not pressed
   '-  bytKey_sto    =1: store key was pressed short (Key_short=1)
   '-  bytKey_sto    =2: store key was pressed long  (Key_long =2)

   If bytKey_press > 0 Then                       'Any key was pressed
      bitKey_up = bytKey_press.Key_up             '=1: Up key pressed
      bitKey_dwn = bytKey_press.Key_dwn           '=1: Down key pressed
      bitKey_enc = bytKey_press.Key_enc           '=1: Encoder key pressed
      bytKey_press.Key_up = 0                     'Reset up key pressed flag
      bytKey_press.Key_dwn = 0                    'Reset down key pressed flag
      bytKey_press.Key_enc = 0                    'Reset encoder key pressed flag
      If bytKey_press.Key_sto = 1 Then            '=1: Store Key pressed
         bytTmp0 = KeyPressedLong(Key_sto)        'Pressed short or long?
         Select Case bytTmp0
            Case 1                                'Store key pressed short
               bytKey_sto = Key_short
            Case 2                                'Store key pressed long
               bytKey_sto = Key_long
         End Select
      Else
         bytKey_sto = 0                           'Store key not pressed
      End If
   End If


   'Enter menu level 2 (settings) from menu 1/2 if key Sto is pressed long
   If bytMenu < 13 And bytKey_sto = Key_Long Then
      bytMenu = 21                                'Menu level 2,1: set SSB IF freq.
      Locate 1 , 1
      LCD "Set SSB IF freq."
      Call PrepareIFCenter(lngIFSSB)
      bytKey_sto = 0
      Cursor ON                                   'Blink: menu levels 2.1 to 2.4
   End If

   If bytMenu = 11 And bitPTT = 1 Then            'Freq. change only in RX mode
      bitTmp0 = bitKey_up XOR bitKey_dwn          'Key Up or down, not both
      If bitTmp0 = 1 Then                         'Change band
         If bitKey_up = 1 Then                    'Increment band
            If bytActBand < bytBands Then
               Incr bytActBand
            Else
               bytActBand = 1
            End If
         End If
         If bitKey_dwn = 1 Then                   'Decrement band
            If bytActBand > 1 Then
               Decr bytActBand
            Else
               bytActBand = bytBands
            End If
         End If
         Call SetGetLastFreq(2)                   'Get last freq. from lngLastFreq
         lngLO = CalcLOFreq(lngFreq)              'LO freq
         Call SetSmoothLimits                     'Set limits smooth tuning
         Call SetNewFrequency                     'New frequency to Si570
         dblFoutOld = dblFout
         Call ShowMHz(lngFreq , 2)
         Call CheckBandLimits(lngFreq)
         Call SetBPFRelais(bytActBand , 1)        'Set BPF relais via I2C
         bitKey_up = 0                            'Reset key up/down flags
         bitKey_dwn = 0
      End If
    End If

   'Check Encoder key, menue 1.1 & 1.2 ----------------------------------------
   If bytMenu < 13 And bitKey_Enc = 1 Then        'Encoder key pressed
      Incr bytMenu
      Call ClearLCDRow(2)

      bitStepSet = 0
      Select Case bytMenu                         'Prepare data for menu
         Case 11                                  'Show band & frequency
            Call ShowMHz(lngFreq , 2)
            Call CheckBandLimits(lngFreq)
            If bitStepSet = 0 Then                'Tuning step not changed
               Call SetGetLastFreq(2)             'Get last freq. from lngLastFreq
               lngFreq = lngBandLo(bytActBand)    'Start freq band
               lngLO = CalcLOFreq(lngFreq)        'Set LO freq
               Call SetBPFRelais(bytActBand , 1)  'Set BPF relais via I2C
               Call SetSmoothLimits               'Set smooth tuning limits
               Call SetNewFrequency               'New frequency to Si570
            End If
         Case 12                                  'Set/change frequency step
            Locate 1 , 1
            LCD "Set freq. step  "
            Locate 2 , 1
            LCD "Step: "
            Locate 2 , 7
            str10 = Str(lngStep)
            str10 = Format(str10 , "000.000")
            Lcd str10 ; " Hz"
         Case 13                                  'from menu level 1.2 to 1.1
            bytMenu = 11
            Call ClearLCDRow(1)
            Call ShowMHz(lngFreq , 2)
      End Select
      bitKey_Enc = 0                              'Reset encoder key flag
   End If

   'Check store key, pressed short, menues 2.2 to 3.0 -------------------------
   If bytMenu > 12 And bytKey_sto = Key_Short Then
      Incr bytMenu
      Call ClearLCDRow(2)
      Select Case bytMenu                         'Prepare data for menu

         Case 22                                  'Set LSB offset
            Locate 1 , 1
            LCD "Set LSB offset  "
            Call PrepareIFOffset(wrdLSBOff , bytLSBSign)

         Case 23                                  'Set USB offset
            Locate 1 , 1
            LCD "Set USB offset  "
            Call PrepareIFOffset(wrdUSBOff , bytUSBSign)

         Case 24                                  'Set CW IF filter center freq
            Locate 1 , 1
            LCD "Set CW IF freq. "
            Call PrepareIFCenter(lngIFCW)

         Case 25                                  'Set CW offset
            Locate 1 , 1
            LCD "Set  CW offset   "
            Call PrepareIFOffset(wrdCWOff , bytCWSign)

         Case 26                                  'Set FM IF filter center freq
            Locate 1 , 1
            LCD "Set FM IF freq. "
            Call PrepareIFCenter(lngIFFM)

         Case 27                                  'Set divider factor
            Locate 1 , 1
            LCD "Set mix divider "
            bytSave = bytLODiv                    'Save divider factor
            Locate 2 , 1                          'Show it
            LCD bytSave
            Cursor off NoBlink

         Case 28                                  'Set Si570 slave address
            Locate 1 , 1
            LCD "Set Si570 addr. "
            bytSave = bytAddrSi570                'Save slave address
            str4 = Hex(bytSave)                   'Show it
            Locate 2 , 1
            LCD "dec " ; Hexval(str4) ; " hex " ; str4

         Case 29                                  'Calibrate Si570 xtal freq.
            Locate 1 , 1
            LCD "Cal Si570 10 MHz"
            lngLO = 10000000                      'LO out = 10 MHz (WWV)
            bitSmooth = 0                         'No smooth tuning
            Call SetNewFrequency                  'New frequency to Si570
            dblSave = dblFxtal                    'Save xtal freq.
            Locate 2 , 1                          'Show it
            dblTmp0 = dblFxtal * 1000000          'Hz
            lngSave = dblTmp0
            Call ShowMHz(lngSave , 2)             'Show xtal frequency
            bitCal = 0
            bytBPFNum = 0
            bytMenuBPF = bytMenu + bytBandsMax    'Last BPF menu number

         Case 30 To bytMenuBPF                    'Set bandpass filter data
            If bytMenu = 30 Then
               Call EEPROMReadBands(1)            'Read all band data from EEPROM
               CLS
            End If
            bytBPFNum = bytBPFNum + 1             'Next filter
            str4 = Str(bytBand(bytBPFNum))        'Show band
            str4 = Format(str4 , "  0")
            Locate 1 , 1
            LCD str4 ; "m "
            bytBPFCount = bytBandBPF(bytBPFNum)
            bitBPFUsed = bytBandUsed(bytBPFNum)
            Locate 1 , 6
            LCD "used=1  " ; bitBPFUsed ; "  "
            Locate 2 , 1
            LCD str4 ; "m"
            str2 = Str(bytBPFCount)
            str2 = Format(str2 , " 0")
            Locate 2 , 6
            LCD "BPFnum " ; str2
            Locate 1 , 14
            bitCal = 0                            '1st Row
            Cursor ON

         Case IS > bytMenuBPF                     'Exit menu
            Call EEPROMReadBands(0)               'Read band data from EEPROM
            CLS
            Call ShowMHz(lngFreq , 2)
            bytMenu = 11                          'Back to menu level 1.1
            dblFxtal = dblSave                    'New or saved xtal frequency
            lngLO = CalcLOFreq(lngFreq)           'LO freq
            Call SetSmoothLimits                  'Set limits smooth tuning
            Call SetNewFrequency                  'New frequency to Si570
            Cursor off NoBlink

      End Select
      bytKey_sto = 0                              'Reset store key flag
   End If


   'Handle dynamic menu data --------------------------------------------------
   Select Case bytMenu

      'Menu 1.1: Set/show TRX freq. -------------------------------------------
      Case 11
         Locate 1 , 1
         str4 = Str(bytBand(bytActBand))          'Show band
         str4 = Format(str4 , "  0")
         LCD str4 ; "m"
         Call TRXData                             'Show TRX data on LCD
         If bitPTT = 1 Then                       'Only RX mode
            dblFoutOld = dblFout                  'LO freq before change
            Disable Timer0                        'Stop Timer0
            Select Case bytIncr                   'Set/show TRX freq.
               Case 2                             'Encoder incremented
                  lngFreq = lngFreq + lngStep
               Case 1                             'Encoder decremented
                  lngFreq = lngFreq - lngStep
            End Select
            'Set new freq if encoder or TRX data are changed
            If bytIncr > 0 Or bitTRXDataChanged = 1 Then
               Call ShowMHz(lngFreq , 2)
               lngLO = CalcLOFreq(lngFreq)        'LO freq
               Call CheckSmoothLimits             'Check smooth tuning limits
               Call SetNewFrequency               'New frequency to Si570
               Call SetGetLastFreq(1)             'Set last freq. in lngLastFreq
               bytIncr = 0
            End If
            Enable Timer0                         'Restart Timer0
         End If
         Call CheckBandLimits(lngFreq)

      'Menu 1.2: Set freq. step -----------------------------------------------
      Case 12
         Disable Timer0                           'Stop Timer0
         Select Case bytIncr
            Case 2                                'Encoder incremented
               If lngStep < 100000 Then lngStep = lngStep * 10
            Case 1                                'Encoder decremented
               If lngStep > 10 Then lngStep = lngStep / 10
         End Select
         If bytIncr > 0 Then                      'Freq. was changed
            Locate 2 , 7
            str10 = Str(lngStep)
            str10 = Format(str10 , "000.000")
            Lcd str10 ; " Hz"
            bytIncr = 0
            bitStepSet = 1
         End If
         Enable Timer0                            'Restart Timer0

      'Menu 2.1: Set SSB IF filter center freq. -------------------------------
      Case 21
         lngSave = SetValue(lngSave)              'Set new IF
         Call ShowMHz(lngSave , 2)                'Show it
         If bytKey_sto = Key_Long Then            'Sto key pressed long
            lngIFSSB = lngSave                    'New IF freq
            lngLO = CalcLOFreq(lngFreq)           'New LO freq
            Locate 2 , 16
            LCD "S"
            wrdAddr = 20                          'EEPROM addr. SSB IF filter freq
            Call EEPROMWriteLng(lngIFSSB)         'Write to EEPROM
            bytKey_sto = 0
         End If
         Locate 2 , bytCursorPos
         Waitms 150

      'Menu 2.2: Set LSB offset -----------------------------------------------
      Case 22
         lngSave = SetValue(lngSave)              'Set new LSB offset
         Call ShowkHz(lngSave , 2)                'Show it
         If bytKey_sto = Key_Long Then            'Sto key pressed long
            wrdLSBOff = lngSave                   'New LSB offset
            bytLSBSign = bytSign                  'new offset sign
            lngLO = CalcLOFreq(lngFreq)           'New LO freq
            Locate 2 , 16
            LCD "S"
            wrdAddr = 24                          'EEPROM addr. LSB offset
            Call EEPROMWriteWrd(wrdLSBOff)        'Write to EEPROM
            wrdAddr = wrdAddr + 2
            WriteEEPROM bytLSBSign , wrdAddr
            bytKey_sto = 0
         End If
         Locate 2 , bytCursorPos
         Waitms 100

      'Menu 2.3: Set USB offset -----------------------------------------------
      Case 23
         lngSave = SetValue(lngSave)              'Set new USB offset
         Call ShowkHz(lngSave , 2)                'Show it
         If bytKey_sto = Key_Long Then            'Sto key pressed long
            wrdUSBOff = lngSave                   'New LSB offset
            bytUSBSign = bytSign                  'new offset sign
            lngLO = CalcLOFreq(lngFreq)           'New LO freq
            Locate 2 , 16
            LCD "S"
            wrdAddr = 27                          'EEPROM addr. LSB offset sign
            Call EEPROMWriteWrd(wrdUSBOff)        'Write to EEPROM
            wrdAddr = wrdAddr + 2
            WriteEEPROM bytUSBSign , wrdAddr
            bytKey_sto = 0                        'Reset store key flag
         End If
         Locate 2 , bytCursorPos
         Waitms 100

      'Menu 2.4: Set CW IF filter center freq. --------------------------------
      Case 24
         lngSave = SetValue(lngSave)              'Set new IF
         Call ShowMHz(lngSave , 2)                'Show it
         If bytKey_sto = Key_Long Then            'Sto key pressed long
            lngIFCW = lngSave                     'New IF freq
            lngLO = CalcLOFreq(lngFreq)           'New LO freq
            Locate 2 , 16
            LCD "S"
            wrdAddr = 30                          'EEPROM addr. CW IF filter freq
            Call EEPROMWriteLng(lngIFCW)          'Write to EEPROM
            bytKey_sto = 0                        'Reset store key flag
         End If
         Locate 2 , bytCursorPos
         Waitms 150

      'Menu 2.5: Set CW offset ------------------------------------------------
      Case 25
         lngSave = SetValue(lngSave)              'Set new USB offset
         Call ShowkHz(lngSave , 2)                'Show it
         If bytKey_sto = Key_Long Then            'Sto key pressed long
            wrdCWOff = lngSave                    'New CW offset
            bytCWSign = bytSign                   'new offset sign
            lngLO = CalcLOFreq(lngFreq)           'New LO freq
            Locate 2 , 16
            LCD "S"
            wrdAddr = 34                          'EEPROM addr. CW offset
            Call EEPROMWriteWrd(wrdCWOff)         'Write to EEPROM
            wrdAddr = wrdAddr + 2
            WriteEEPROM bytCWSign , wrdAddr
            bytKey_sto = 0                        'Reset store key flag
         End If
         Locate 2 , bytCursorPos
         Waitms 100

      'Menu 2.6: Set FM IF filter center freq. --------------------------------
      Case 26
         lngSave = SetValue(lngSave)              'Set new IF
         Call ShowMHz(lngSave , 2)                'Show it
         If bytKey_sto = Key_Long Then            'Sto key pressed long
            lngIFFM = lngSave                     'New IF freq
            lngLO = CalcLOFreq(lngFreq)           'New LO freq
            Locate 2 , 16
            LCD "S"
            wrdAddr = 37                          'EEPROM addr. FM IF filter freq
            Call EEPROMWriteLng(lngIFCW)          'Write to EEPROM
            bytKey_sto = 0                        'Reset store key flag
         End If
         Locate 2 , bytCursorPos
         Waitms 150

      'Menu 2.7: Set LO divider -----------------------------------------------
      Case 27
         Disable Timer0                           'Stop Timer0
         Select Case bytIncr
            Case 2                                'Encoder incremented
               If bytSave < 8 Then Shift bytSave , Left , 1
            Case 1                                'Encoder decremented
               If bytSave >= 2 Then Shift bytSave , Right , 1
         End Select
         If bytSave = 0 Then bytSave = 1
         If bytIncr > 0 Then                      'Show changed value
            Locate 2 , 1
            LCD bytSave
         End If
         bytIncr = 0                              'Reset encoder increment
         If bytKey_sto = Key_Long Then            'Sto key pressed long
            bytLODiv = bytSave                    'Save divider factor
            lngLO = CalcLOFreq(lngFreq)           'New LO frequency
            Locate 2 , 16
            LCD "S"
            wrdAddr = 9                           'EEPROM addr. LO divider                          '
            WriteEEPROM bytLODiv , wrdAddr        'Write to EEPROM
            bytKey_sto = 0                        'Reset store key flag
         End If
         Enable Timer0                            'Restart Timer0

      'Menu 2.8: Set Si570 slave address --------------------------------------
      Case 28
         Disable Timer0                           'Stop Timer0
         Select Case bytIncr
            Case 2                                'Encoder incremented
               If bytSave < 255 Then Incr bytSave
            Case 1                                'Encoder decremented
               If bytSave > 1 Then Decr bytSave
         End Select
         If bytIncr > 0 Then                      'Show changed value
            str4 = Hex(bytSave)
            Locate 2 , 1
            LCD "dec " ; Hexval(str4) ; " hex " ; str4
         End If
         bytIncr = 0                              'Reset encoder increment flag
         If bytKey_sto = Key_Long Then            'Sto key pressed long
            bytAddrSi570 = bytSave                'Save slave address
            Call CalcSi570Addr                    'recalculate addresses
            Locate 2 , 16
            LCD "S"
            wrdAddr = 10                          'EEPROM addr.Si570 slave address
            WriteEEPROM bytAddrSi570 , wrdAddr    'Write to EEPROM
            bytKey_sto = 0                        'Reset store key flag
         End If
         Enable Timer0                            'Restart Timer0

      'Menu 2.9: Calibrate Si570 xtal, LO out 10 MHz --------------------------
      'bitCal  = 0: set/show xtal freq.
      '        = 1: set freq. step
      Case 29
         If bitKey_Enc = 1 Then                   'Encoder key pressed
            If bitCal = 0 Then                    'Enter freq. step mode
               Locate 1 , 1
               LCD "Set freq. step  "
               Locate 2 , 1
               LCD "Step: "
               Locate 2 , 7
               str10 = Str(lngStepCal)
               str10 = Format(str10 , "000.000")
               Lcd str10 ; " Hz"
               bitCal = 1                         'Now in freq. step mode
            Else                                  'Enter xtal calibration mode
               Locate 1 , 1
               LCD "Cal Si570 10 MHz"
               lngLO = 10000000                   'LO out = 10 MHz (WWV)
               bitSmooth = 0                      'No smooth tuning
               Call SetNewFrequency               'New frequency to Si570
               dblSave = dblFxtal                 'Save xtal freq. [MHz]
               Locate 2 , 15
               LCD "  "
               Locate 2 , 1                       'Show it
               Call ShowMHz(lngFxtal , 2)         'Show xtal frequency [Hz]
               bitCal = 0                         'Now in xtal calibration mode
            End If
            bitKey_Enc = 0                        'Reset encoder key flag
         End If
         Disable Timer0                           'Stop Timer0
         If bitCal = 1 Then                       'change freq. step
            Select Case bytIncr
               Case 2                             'Encoder incremented
                    If lngStepCal < 100000 Then lngStepCal = lngStepCal * 10
               Case 1                             'Encoder decremented
                    If lngStepCal > 10 Then lngStepCal = lngStepCal / 10
            End Select
            If bytIncr > 0 Then                   'Step freq. was changed
               Locate 2 , 7                       'Show step freq. [Hz]
               str10 = Str(lngStepCal)
               str10 = Format(str10 , "000.000")
               Lcd str10 ; " Hz"
               bytIncr = 0                        'Reset encoder increment flag
            End If
         Else                                     'change xtal freq.
            Select Case bytIncr
               Case 2                             'Encoder incremented
                  lngFxtal = lngFxtal + lngStepCal
               Case 1                             'Encoder decremented
                  lngFxtal = lngFxtal - lngStepCal
            End Select
            If bytIncr > 0 Then                   'Show changed value
               Locate 2 , 1
               dblFxtal = lngFxtal                'Hz
               dblFxtal = dblFxtal * 0.000001     'MHz for Si570 calculation
               Call ShowMHz(lngFxtal , 2)         'Show xtal frequency
               Call SetNewFrequency               'New frequency to Si570
               bytIncr = 0                        'Reset encoder increment flag
            End If
            If bytKey_sto = Key_Long Then         'Sto key pressed long
               dblSave = dblFxtal                 'New xtal frequency [MHz]
               Locate 2 , 16
               LCD "S"
               wrdAddr = 11                       'EEPROM addr. Si570 xtal freq.
               Call EEPROMWriteLng(lngFxtal)      'Write to EEPROM [Hz]
               bytKey_sto = 0                     'Reset store key flag
            End If
         End If                                   'End of mode 29
         Enable Timer0                            'Restart Timer0

      'Menu 3.0: Set bandpass filter data ------------------------------------
      'Row 1: BPF used = 1 / not used = 0
      'Row 2: BPF sequential number
      'Toggle row1/2 with encoder key
      Case 30 To bytMenuBPF
         If bitKey_Enc = 1 Then                   'Encoder key pressed
            If bitCal = 0 Then                    '1st row, BPF used or not
               Locate 2 , 14
               bitCal = 1
            Else                                  '2nd row, BPF number
               Locate 1 , 14
               bitCal = 0
            End If
            bitKey_Enc = 0                        'Reset encoder increment flag
         End If
         If bitCal = 0 Then                       '1st row, BPF used or not
            If bytIncr > 0 Then                   'Encoder was turned
               Toggle bitBPFUsed
               Locate 1 , 14
               LCD bitBPFUsed
               Locate 1 , 14
               bytIncr = 0                        'Reset encoder increment flag
            End If
         Else                                     '2nd row, BPF number
            Disable Timer0                        'Stop Timer0
            Select Case bytIncr
               Case 2                             'Encoder incremented
                  Incr bytBPFCount
                  If bytBPFCount > bytBandsMax Then
                     bytBPFCount = bytBandsMax
                  End If
               Case 1                             'Encoder decremented
                  Decr bytBPFCount
                  If bytBPFCount < 1 Then
                     bytBPFCount = 1
                  End If
            End Select
            If bytIncr > 0 Then
               str2 = Str(bytBPFCount)
               str2 = Format(str2 , " 0")
               Locate 2 , 13
               LCD str2
               Locate 2 , 14
               bytIncr = 0                        'Reset encoder increment flag
            End If
            Enable Timer0                         'Restart Timer0
         End If
         If bytKey_sto = Key_Long Then            'Sto key pressed long
            Locate 2 , 16                         'Write data to EEPROM
            LCD "S"
            wrdAddr = bytBPFNum - 1
            wrdAddr = 11 * wrdAddr                '11 bytes per band
            wrdAddr = wrdAddr + 70                '1st BPF EEPROM address
            bytTmp1 = bitBPFUsed
            WriteEEPROM bytTmp1 , wrdAddr
            wrdAddr = wrdAddr + 2                 'EEPROM Address BPF number
            WriteEEPROM bytBPFCount , wrdAddr
            bytKey_sto = 0                        'Reset store key flag
            Wait 1
         End If

   End Select

Loop

End

'End of main program **********************************************************

'==============================================================================
Timer0_isr:
'Encoder interrupt on Timer0 overflow
'While data are displayed Timer0 should be stopped
'Output: bytIncr = 2: increment, =1: decrement if encoder was changed
'Having processed the bytIncr data in the main program reset bytIncr = 0
'Switching pattern of ALPS STEC11 encoder:
'Turning clockwise              ... counter clockwise
'A    B                             A     B
'1    1  start detent position      1     1
'1    0  intermediate position 1    0     1
'0    0  intermediate position 2    0     0
'0    1  intermediate position 3    1     0
'1    1  next detent position       1     1

   Timer0 = PresetTimer0

   'Encoder -------------------------------------------------------------------
   bytEnc_new.0 = Enc_A                           'Set bit 0 of bytEnc_new
   bytEnc_new.1 = Enc_B                           'Set bit 1 of bytEnc_new

   'Check encoder pattern if encoder is changed
   'bytIncr = 1: decremented (turned counter clockwise)
   'bytIncr = 2: incremented (turned         clockwise)
   If bytEnc_new <> bytEnc_old Then               'encoder is changed
      'from intermediate position 3 to quiescent position
      If bytEnc_new = &B00000011 And bytEnc_old = &B00000010 Then bytIncr = 2
      If bytEnc_new = &B00000011 And bytEnc_old = &B00000001 Then bytIncr = 1
      'from intermediate position 1 to intermediate position 2
      'DoubleStep = 1:  intermediate position 2 gives an additional tick
      '                 2 ticks between 2 detent positions on slow rotation
      '           = 0:  1 tick  between 2 detent positions
      'DoubleStep is set on the front panel board:
      'DoubleStep = 0: Connect front panel board contact strip (solder bridge)
      '                pin 20 to adjacent GND pin, is the former "Spare" pin
      'DoubleStep = 1: Nothing to do

      If Pin_DoubleStep = 1 Then
         If bytEnc_new = &B00000000 And bytEnc_old = &B00000001 Then bytIncr = 2
         If bytEnc_new = &B00000000 And bytEnc_old = &B00000010 Then bytIncr = 1
      EndIf

      bytEnc_old = bytEnc_new                     'old <- new for the next time
   End If

   'Keys ----------------------------------------------------------------------
   'got from Thomas Baer in microcontroller.net
   'Original (ASM) by Peter Dannegger
   'http://www.mikrocontroller.net/topic/tasten-entprellen-bulletproof

   'Input:  Key_port       Each of the 8 bits of byte variable represents one key
   '                       bits 0...7: keys 0...7 @ Key_port
   '        bytKey_mask    Bit pattern to mask the keys enabled for short/long
   '                       Key number (0...7) to be checked for short/long
   '                       e.g. "00000101" Keys 0 and 2 are short/long keys
   'Output: bytKey_press   bit pattern, key which changed from open to pressed
   '                       Reset after evaluation in the main program
   '                       Bits remain = 1 until resetted in the main program
   '        bytKey_state   bit pattern, Key state, 1:pressed or 0:not pressed
   '                       Bits = 1 while key is pressed, = 0: Key is released
   '        bytKey_save    last state of bytKey_state before entering Timer0_isr
   '                       So it can be determined wether between two calls
   '                       of Timer0_isr any key was pressed.

   'Keys at different ports
   Key_port.Key_sto = Pin_KeySto                  'Key store
   Key_port.Key_up = Pin_KeyUp                    'Key up
   Key_port.Key_dwn = Pin_Keydwn                  'Key down
   Key_port.Key_enc = Pin_KeyEnc                  'Key encoder


   bytKey_save = bytKey_state                     'save old status before keys were pressed
   bytIwr0 = Key_port                             'get Bits of the KeyPort
   bytIwr0 = Not bytIwr0
   bytIwr0 = bytIwr0 Xor bytKey_state
   bytKey_ct0 = bytKey_ct0 And bytIwr0
   bytKey_ct1 = bytKey_ct1 And bytIwr0
   bytKey_ct0 = Not bytKey_ct0
   bytKey_ct1 = bytKey_ct1 Xor bytKey_ct0
   bytIwr0 = bytIwr0 And bytKey_ct0
   bytIwr0 = bytIwr0 And bytKey_ct1
   bytKey_state = bytKey_state Xor bytIwr0
   bytIwr0 = bytIwr0 And bytKey_state
   bytKey_press = bytKey_press Or bytIwr0

   'Counter for short/long detection
   bytIwr0 = bytKey_state AND bytKey_mask         'Mask out short/long keys
   If bytIwr0 > 0 Then                            'Short/long key still pressed
      Incr wrdTimer0_runs                         'Clock is running...
   End If
Return

'==============================================================================
Function KeyPressedLong(byVal bytKey_Long As Byte) As Byte
   'Check if key bytKey_Long (0...7) in Key_port is pressed short or long
   'Output: KeyPressedLong =  0: evaluating, key is still pressed
   '                          1: key is released, pressed short
   '                          2: key is released, pressed long

   If bytKey_state.bytKey_Long = 1 Then           'Key still pressed
      KeyPressedLong = 0
   Else                                           'Key released
      If wrdTimer0_runs < wrdTimer0_long Then     'short time elapsed
         KeyPressedLong = 1                       'Pressed short
      Else
         KeyPressedLong = 2                       'Pressed short
      End If
      wrdTimer0_runs = 0
      bytKey_press.bytKey_Long = 0                'Reset pressed flag
   End If
End Function

'==============================================================================
Sub ClearLCDRow(byVal bytRow As Byte)
'Clear LCD Row bytRow
   Locate bytRow , 1
   LCD SPC(16)
   Locate bytRow , 1
End Sub

'==============================================================================
Sub ShowMHz(ByVal lngF As Long , byVal bytRow As Byte)
'show frequency lngF on LCD

   str10 = Str(lngF)
   str10 = Format(str10 , "  0.000000")
   str11 = Left(str10 , 3) + "," + Mid(str10 , 5 , 3) + "," + Right(str10 , 3)
   Locate bytRow , 1
   Lcd str11 ; " Hz"
End Sub

'==============================================================================
Sub SetGetLastFreq(ByVal bytOpt As Byte)
'Set or get last frequency used in band
'Input:  bytOpt     = 1: set current freq. in array lngLastFreq
'        lngFreq    = current frequency
'
'        bytOpt     = 2: get last freq.  from array lngLastFreq
'        lngLastFreq= last freq, used in current band
'        bytActBand = current band index

   Select Case bytOpt
      Case 1                                      'Set current freq. lngFreq in array lngLastFreq
         lngLastFreq(bytActBand) = lngFreq
      Case 2                                      'get last freq. from array lngLastFreq
         If lngLastFreq(bytActBand) = 0 Then      'Not yet initialized
            lngFreq = lngBandLO(bytActBand)       'Band start TRX freq
         Else                                     'Set last frequency used
            lngFreq = lngLastFreq(bytActBand)
         End If
   End Select
End Sub

'==============================================================================
Sub CheckBandLimits(byVal lngF As Long)
'Check TRX frequency with respect to band limits
'SSB Limits are enlarged by +/- 3 kHz for security

   bytTmp0 = 0
   lngTmp0 = lngBandLo(bytActBand)
   lngTmp1 = lngBandHi(bytActBand)
   If bitFM = 1 Then                              'FM not set: SSB or CW mode
      If bitSSB = 1 Then                          'SSB mode
         If bitLSB = 1 Then                       'LSB: increase lower limit
            lngTmp0 = lngTmp0 + 1500              '~ half SSB bandwidth
         Else                                     'USB: decrease upper limit
            lngTmp1 = lngTmp1 - 1500
         End If
      End If
   Else                                           'FM mode
      lngTmp0 = lngTmp0 + 6000                    '~ half FM bandwidth
      lngTmp1 = lngTmp1 - 6000
   End If
   If lngFreq < lngTmp0 Then                      'Check lower band limit
      bytTmp0 = 1
   End If
   If lngFreq > lngTmp1 Then                      'Check upper band limit
      bytTmp0 = 2
   End If

   Locate 1 , 16
   Select Case bytTmp0
      Case 0
         LCD " "                                  'Within limits
      Case 1
         LCD "<"                                  'Below lower limt
      Case 2
         LCD ">"                                  'Above upper limit
   End Select
End Sub

'==============================================================================
Sub ShowkHz(ByVal lngF As Long , byVal bytRow As Byte)
'show frequency lngF on LCD

   Locate bytRow , bytSignPos
   Select Case bytSign
      Case 0                                      'No sign
         LCD " "
      Case 1                                      'Subtract offset
         LCD "-"
      Case 2
         LCD "+"                                  'Add offset
   End Select
   str10 = Str(lngF)
   str10 = Format(str10 , "  0.000")
   str11 = Left(str10 , 3) + "," + Right(str10 , 3)
   bytTmp0 = bytSignPos + 1
   Locate bytRow , bytTmp0
   Lcd str11 ; " Hz"
End Sub

'==============================================================================
Sub PrepareIFOffset(byVal wrdOffFreq As Word , byVal bytOffSign As Byte)
'Prepare data for IF filter offset setting
'Input   wrdOffFreq  Offset frequency
'        bytOffSign  Offset frequency sign (+/-)

   lngSave = wrdOffFreq
   bytCursorPos = GetCursorPos(lngSave)
   bytCursorMin = 4                               'First LCD cursor position
   bytCursorMax = 11                              'Last  LCD cursor position
   bytNumChar = 6                                 'kHz: 6 characters
   bytSignPos = 4                                 'LCD Position +/- sign
   bytSign = bytOffSign
   Call ShowkHz(lngSave , 2)
End Sub

'==============================================================================
Sub PrepareIFCenter(byVal lngIFFreq As Long)
'Prepare data for IF filter center frequency setting
'Input   lngIFFreq   IF filter center frequency

   lngSave = lngIFFreq
   lngFreqSave = lngFreq
   bytCursorPos = GetCursorPos(lngSave)
   bytCursorMin = 1                               'First LCD cursor position
   bytCursorMax = 11                              'Last  LCD cursor position
   bytNumChar = 9                                 'MHz: 9 characters
   bytSignPos = 0                                 'LCD Position +/- sign (no sign)
   bytSign = 0                                    'No sign
   Call ShowMHz(lngSave , 2)
End Sub

'==============================================================================
Function GetCursorPos(lngF As Long) As Byte
'Calculate the cursor position (most left position) of freq. lngF
'lngF on LCD: 000,000,000 (Hz)
'Input:  lngF, frequency (Hz)
'Output: GetCursorPos: position of most left digit

   str10 = str(lngF)
   intTmp0 = Len(str10)                           'Length without decimal ","
   GetCursorPos = 10 - intTmp0                    'Left "000" pack
   Select Case intTmp0
      Case 7 To 9                                 'Left "000" pack
         'Nothing to do
      Case 4 To 6                                 'Mid "000" pack
         GetCursorPos = GetCursorPos + 1
      Case 1 To 3                                 'Right "000" pack
         GetCursorPos = GetCursorPos + 2
   End Select
End Function

'==============================================================================
Function SetValue(byVal lngVal As Long) As Long
'Set value lngVal via encoder
'Digits are selected via up/down keys
'Input:  lngVal         long integer value to be set
'        bytCursorPos   Cursor position (digit)
'        bitKey_up      =1: key up was pressed
'        bitKey_dwn     =1: key down was pressed
'Output: SetValue       changed lngVal

   str10 = Str(lngVal)                            'Frequency before change
   str10 = Format(str10 , "000000000")            'Fill "0" on the left

   If bitKey_dwn = 1 Then                         'Select digit >> downwards
      If bytCursorPos < bytCursorMax Then
         Incr bytCursorPos
         If bytNumChar = 9 Then
            If bytCursorPos = 4 Or bytCursorPos = 8 Then       'Dec. points MHz
               Incr bytCursorPos
            End If
         Else
            If bytCursorPos = 8 Then              'Dec. point kHz
               Incr bytCursorPos
            End If
         End If
      End If
      bitKey_dwn = 0
   End If
   If bitKey_up = 1 Then                          'Select digit << upwards
      If bytCursorPos > bytCursorMin Then
         Decr bytCursorPos
         If bytNumChar = 9 Then
            If bytCursorPos = 4 Or bytCursorPos = 8 Then
               Decr bytCursorPos
            End If
         Else
            If bytCursorPos = 8 Then
               Decr bytCursorPos
            End If
         End If
      End If
      bitKey_up = 0
   End If

   bytTmp0 = bytCursorPos                         'Get value @ cursor position
   Select Case bytCursorPos                       'Cancel the "," on LCD
      Case 5 To 7
         bytTmp0 = bytTmp0 - 1                    'middle "000" pack
      Case 9 To 11
         bytTmp0 = bytTmp0 - 2                    'right "000" pack
   End Select
   If bytCursorPos <> bytSignPos Then
      str1 = Mid(str10 , bytTmp0 , 1)             'Value to be changed
      bytSave = Val(str1)
   Else
      bytSave = 0
   End If

   Select Case bytIncr                            'Set digit via encoder
      Case 2                                      'Encoder incremented
         If bytCursorPos = bytSignPos Then        'Sign for kHz
            If bytSign = 1 Then bytSign = 2       '- to +
         Else
            If bytSave < 9 Then
               Incr bytSave
            Else
               bytSave = 0
            End If
         End If
      Case 1                                      'Encoder decremented
         If bytCursorPos = bytSignPos Then        'Sign for kHz
            If bytSign = 2 Then bytSign = 1       '+ to -
         Else
            If bytSave > 0 Then
               Decr bytSave
            Else
               bytSave = 9
            End If
         End If
   End Select
   If bytIncr > 0 Then                            'Digit was changed
      Locate 2 , bytCursorPos                     'Show changed digit
      LCD bytSave
      SetValue = SetIFfreq(lngVal , bytSave)
      bytIncr = 0
   End If
End Function

'==============================================================================
Function SetIFfreq(byVal lngF As Long , byVal bytDigit As Byte) As Long
'Set IF or offset frequency with encoder
'Input:  lngF IF freq before change
'        bytDigit: changed digit
'        bytCursorPos, position of changed digit on LCD
'        LCD formatting:   000,000,000 for IF, MHz
'                          xxxs000,000 for offset, kHz (xxx=blank, s=sign)
'Output: SetIFfreq, new frequency (long integer)

   str10 = Str(lngF)                              'IF freq before change
   intTmp0 = Len(str10)
   str10 = Format(str10 , "000000000")            'Fill "0" on the left
   str1 = str(bytDigit)                           'digit to be put into str10

   bytTmp0 = bytCursorPos
   Select Case bytCursorPos                       'Cancel the "," on LCD
      Case 5 To 7
         bytTmp0 = bytTmp0 - 1                    'middle "000" pack
      Case 9 To 11
         bytTmp0 = bytTmp0 - 2                    'right "000" pack
   End Select
   If bytTmp0 = bytCursorMin Then                 'put str1 into string
      str11 = str1 + Right(str10 , 8)
   Else
      intTmp0 = bytTmp0 - 1
      intTmp1 = 9 - bytTmp0
      str11 = Left(str10 , intTmp0) + str1 + Right(str10 , intTmp1)
   End If
   SetIFfreq = Val(str11)                         'String to long integer
End Function

'==============================================================================
Function CheckI2C(byVal bytAdr As Byte) As Byte
'Check I2C chip @ address bytAdr
'CheckI2C = 1: OK, 0: else

   I2CStart
   I2CwByte bytAdr                                'Try to write I2C address
   I2CStop
   If Err = 1 Then                                'Error raised
      CheckI2C = 0                                'Not OK
   Else
      CheckI2C = 1                                'Is OK
   End If
End Function

'==============================================================================
Sub WriteI2C(byVal bytAdr As Byte , bytByte As Byte)
'Write Byte bytByte to I2C chip @ address bytAdr

   I2CStart
   I2CwByte bytAdr                                'Write I2C address
   I2CwByte bytByte                               'Write Byte
   I2CStop
End Sub

'==============================================================================
Sub SetBPFRelais(byVal bytBandNum As Byte , byVal bytStat As Byte)
'Set Bandpass filter relais status via I2C
'Max 16 BPF (1...16) for 2 PCF8574A
'Bank 1: (1... 8), first  PCF8574A
'Bank 2: (9...16), second PCF8574A
'Input:  bytBandNum              Number of band (1...max 16)
'        bytBandBPF(bytBandNum)  Number of bandpass filter for this band
'        bytStat     Status, On=1, off=0

   bytTmp2 = bytBandBPF(bytBandNum)               'Bandpass filter number

   Select Case bytTmp2
      Case 1 To 8                                 'Relais Bank 1
         If bytBPFs > 8 Then                      '2 banks used
            bytTmp0 = 0
            Call WriteI2C(bytI2C_BPF2 , bytTmp0)  'Clear Bank 2
         End If
         bytTmp0 = bytTmp2 - 1                    'Count: 0, 1,...
         bytTmp1 = bytI2C_BPF1                    'Addr. 1st PCF8574A
      Case 9 To 16                                'Relais Bank 2
         bytTmp0 = 0
         Call WriteI2C(bytI2C_BPF1 , bytTmp0)     'Clear Bank 1
         bytTmp0 = bytTmp2 - 9                    'Count: 0, 1,...
         bytTmp1 = bytI2C_BPF2                    'Addr. 2nd PCF8574A
   End Select

   bytBPFStatus = 0                               'Clear status
   bytBPFStatus.bytTmp0 = bytStat                 'Set status
   Call WriteI2C(bytTmp1 , bytBPFStatus)          'Now set relais
End Sub

'==============================================================================
Sub TRXData
'Check TRX data (PTT, CW/SSB, LSB/USB) and show in on LCD
'Set center frequency for SSB-/CW-/FM IF Filter (may be different)
'Port data are set by NPN transistors
'Base voltage = low (not set): Port is high (1) and vice versa
'Input:  Port settings  bitPPT   =1: RX,        =0: TX
'                       bitSSB   =1: SSB,       =0: CW
'                       bitLSB   =1: SSB/LSB,   =0; SSB/USB
'                       bitFM    =1: SSB or CW  =0: FM

   Locate 1 , 6
   If bitPTT = 1 Then                             'PTT not set: RX
      Locate 1 , 6
      LCD "RX"
   Else
      LCD "TX"
   End If
   Locate 1 , 9
   If bitFM = 1 Then                              'FM not set: SSB or CW mode
      If bitSSB = 1 Then                          'SSB not set: SSB mode
         lngIFFreq = lngIFSSB                     'SSB filter freq.
         If bitLSB = 1 Then                       'USB not set: LSB
            LCD "LSB"
         Else
            LCD "USB"
         End If
      Else                                        'CW mode
         lngIFFreq = lngIFCW                      'CW filter freq.
         LCD " CW"
      End If
   Else                                           'FM ist set: FM mode
      lngIFFreq = lngIFFM                         'FM filter freq.
      LCD " FM"
   End If

   'Check if TRX data are changed
   bitTRXDataChanged = 0
   If bitPTT <> bitPTTOld Then                    'RX-/TX mode was changed
      bitTRXDataChanged = 1
      bitPTTOld = bitPTT
   End If
   If bitSSB <> bitSSBOld Then                    'SSB-/CW mode was changed
      bitTRXDataChanged = 1
      bitSSBOld = bitSSB
   End If
   If bitLSB <> bitLSBOld Then                    'LSB-/USB mode was changed
      bitTRXDataChanged = 1
      bitLSBOld = bitLSB
   End If
   If bitFM <> bitFMOld Then                      'FM-/SSB/CW mode was changed
      bitTRXDataChanged = 1
      bitFMOld = bitFM
   End If
End Sub

'==============================================================================
Function CalcLOFreq(ByVal lngTRX As Long) As Long
'Calculate LO freq. from TRX & IF freq. & BFO offset & divider factor
'LO freq = (TRX freq + IF freq +/- BFO offset) * divider factor
'FM mode: no BFO offset
'BFO offset for SSB, RX+TX mode and for CW, RX mode only
'Input:  lngTRX      TRX frequency [Hz]
'        lngIFFreq   IF center freq., current IF filter (SSB/CW/FM) [Hz]
'        bitSSB      =1: SSB,       =0: CW
'        bitLSB      =1: LSB,       =0: USB
'        bit FM      =1: SSB or CW, =0: FM
'        wrdLSBOff   LSB offset
'        wrdUSBOff   USB offset
'        wrdCWOff    CW offset
'        bytLSBSign  sign of LSB offset
'        bytUSBSign  sign of USB offset
'        bytCWSign   sign of CW offset
'        bytLODiv    mixer divider factor
'Output: CalcLOFreq  LO frequency

   lngTmp0 = lngTRX + lngIFFreq                   'LO freq. above TRX freq.
   intTmp0 = 0
   bytTmp0 = 0
   If bitFM = 1 Then                              'FM is not set: SSB or CW
      If bitSSB = 1 Then                          'Mode SSB
         If bitLSB = 1 Then                       'SSB, LSB
            intTmp0 = wrdLSBOff
            bytTmp0 = bytLSBSign                  '1=Subtract Offset, 2=add
         Else                                     'SSB, USB
            intTmp0 = wrdUSBOff
            bytTmp0 = bytUSBSign                  '1=Subtract Offset, 2=add
         End If
      Else                                        'Mode CW
         If bitPTT = 1 Then                       'PTT not set: RX
            intTmp0 = wrdCWOff
            bytTmp0 = bytCWSign                   '1=Subtract Offset, 2=add
         Else                                     'PTT set: TX
            intTmp0 = 0                           'No CW offset in TX mode
            bytTmp0 = 2
         End If
      End If
   End If
   If bytTmp0 = 1 Then                            'Substract offset
      lngTmp0 = lngTmp0 - intTmp0
   Else                                           'Add offset
      lngTmp0 = lngTmp0 + intTmp0
   End If
   If bytLODiv > 1 Then
      bytTmp0 = bytLODiv                          'Divider 2, 4, 8
      Shift bytTmp0 , Right , 1                   '=1, 2, 3
      Shift lngTmp0 , Left , bytTmp0              'multiply with 2, 4, 8
   End If
   CalcLOFreq = lngTmp0
End Function

'==============================================================================
Sub SetNewFrequency
'Set new output frequency in Si570
'Sid Knox, W7QJQ, updated for smooth tuning, G. Latzel
'Input:  lngLO       LO frequency [Hz]
'        bitSmooth   =1, smooth tuning

   If lngLO > 5000000 Then                        '> 5 MHz (spec. 10 MHz min.)
      dblFout = lngLO
      dblFout = dblFout * 0.000001                'Freq [MHz]
      If bitSmooth = 0 Then                       'Full reprogramming
         bytN1 = 1                                'Initialize N1 and HS_DIV
         bytArrayIndex = 1
         Call CalcConstant                        'calc optimum HS_DIV and N1
      End If
      Call CalcRFREQ                              'calculate  RFREQ
      Call PackRegData                            'format and pack the data
      Call WriteSi570regs                         'send it to the Si570
   End If
End Sub

'==============================================================================
Sub CalcConstant()
'Calculate the optimum values for HS_DIV and N1.
'Sid Knox, W7QJQ
'The Si570 data sheet states "The sets of dividers should be sorted
'to minimize Fosc for power dissipation and to minimize N1's power comsumption".

  lngFout = dblFout

Calc:
  lngConstant = bytHS_DIV(bytArrayIndex) * bytN1
  lngCheckFreq = lngFout * lngConstant

  If lngCheckFreq < 4850 Then                     '< min DCO freq [GHz]?
    Goto IncrN1
  Else
    If lngCheckFreq > 5670 Then                   '> max DCO freq [GHz]?
      Goto DecrHS_DIV                             'if too high, go decr HS_DIV
    End If
    bytFinalHS_DIV = bytHS_DIV(bytArrayIndex)     'now have HS_DIV and N1
    Exit Sub                                      'good constant, exit to main
  End If

IncrN1:
  If bytN1 = 1 Then
    Incr bytN1                                    'make it 2
  Else
    Incr bytN1                                    'try next EVEN Higher N1
    Incr bytN1
  End If

  If bytN1 > 126 Then                             'check bounds
    Goto DecrHS_DIV
  End If
  Goto Calc                                       'try again with new N1

DecrHS_DIV:
   Incr bytArrayIndex                             'try next Higher N1
   If bytArrayIndex > 6 Then                      'check bounds
      Locate 2 , 1                                'Show error message
      LCD "No valid freq!  "
      Wait 1
      Stop                                        'Halt program after error msg
   End If
   bytN1 = 1
   Goto Calc                                      'go try again with new HS_DIV
End Sub

'==============================================================================
Sub CalcRFREQ()
'Calculate the tuning word (RFREQ) for the Si570 DCO,
'compute the INTeger and FRACtional parts, and scale the FRAC part by 2^28
'Sid Knox, W7QJQ, updated for smooth tuning, G. Latzel

   If bitSmooth = 0 Then                          'Full programming of DCO
      lngTmp0 = bytHS_DIV(bytArrayIndex) * bytN1  'divisor constant
      dblHS_DIVXN1 = lngTmp0
      dblRFREQ = dblFout * dblHS_DIVXN1           'DCO control word
      dblRFREQ = dblRFREQ / dblFxtal
   Else                                           'Smooth tuning
      dblRFREQ = dblRFREQ * dblFout
      dblRFREQ = dblRFREQ / dblFoutOld
   End If
   dbIintRFREQ = Int(dblRFREQ)                    'make into INT and FRAC parts
   dblFracRFREQ = Frac(dblRFREQ)
   dblFracRFREQ = dblFracRFREQ * dbl2pwr28        'scale times 2^28
End Sub

'==============================================================================
Sub PackRegData()
'Format and pack the 6 frequency control bytes using HS_DIV, N1 and RFREQ
'Sid Knox, W7QJQ, updated for smooth tuning, G. Latzel

'pack Si570 reg 7
   If bitSmooth = 0 Then                          'Full programming of DCO
      bytTmp0 = bytFinalHS_DIV - 4
      Shift bytTmp0 , Left , 5
      bytTmp1 = bytN1 - 1
      Shift bytTmp1 , Right , 2
      bytWriteRegs(1) = bytTmp0 Or bytTmp1        'HS_DIV and partial N1
   End If

'pack Si570 reg 8
   bytTmp0 = bytN1 - 1
   Shift bytTmp0 , Left , 6                       'two msbits N1
   lngTmp0 = dbIintRFREQ
   Shift lngTmp0 , Right , 4
   lngTmp0 = lngTmp0 And 15
   bytWriteRegs(2) = bytTmp0 Or lngTmp0


'pack Si570 reg 9
   lngTmp0 = dbIintRFREQ
   lngTmp0 = lngTmp0 And 15
   Shift lngTmp0 , Left , 4
   lngTmp1 = dblFracRFREQ                         'this is a 24 bit word!!
   Shift lngTmp1 , Right , 24
   bytWriteRegs(3) = lngTmp0 Or lngTmp1

'pack Si570 reg 10
   lngTmp0 = dblFracRFREQ
   Shift lngTmp0 , Right , 16
   bytWriteRegs(4) = lngTmp0

'pack Si570 reg 11
   lngTmp0 = dblFracRFREQ
   Shift lngTmp0 , Right , 8                      '
   bytWriteRegs(5) = lngTmp0

'pack Si570 reg 12
   lngTmp0 = dblFracRFREQ
   bytWriteRegs(6) = lngTmp0
End Sub

'==============================================================================
Sub WriteSi570regs()
'Send the 48-bit tuning word (6-bytes) via I2C.
'Sid Knox, W7QJQ, updated for smooth tuning, G. Latzel
'bitSmooth = 0:   Large changes in output frequency
'                 Output frequency is fully reconfigured
'                 Registers 7 to 12 are set
'                 Freeze (stop) & unfreeze DCO (makes noise)
'bitSmooth = 1:   small changes in output frequency <= 3,500 ppm
'                 Smooth tuning
'                 Only registers 8 to 12 are set
'                 No freeze/unfreeze/assert of DCO (noiseless)

   If bitSmooth = 0 Then                          'Full programming of DCO
      bytTmp1 = 1                                 'Register 7...12 (index 1...6)
      I2cstart                                    'Freeze the DCO
      I2cwbyte bytSi570WriteAdr
      I2cwbyte 137                                'set bit4, reg 137
      I2cwbyte &H10
      I2cstop                                     'Freeze finished
   Else
      bytTmp1 = 2                                 'Register 8...12 (index 2...6)
   End If
   bytTmp2 = bytTmp1 + 6                          'Register address
   I2cstart
   I2cwbyte bytSi570WriteAdr                      'Write address
   I2cwbyte bytTmp2                               'Write register address
   For bytTmp0 = bytTmp1 To 6                     'write register bytes
      I2cwbyte bytWriteRegs(bytTmp0)
   Next bytTmp0
   I2cstop
   If bitSmooth = 0 Then                          'Full programming of DCO
      I2cstart                                    'UnFreeze the DCO
      I2cwbyte bytSi570WriteAdr
      I2cwbyte 137                                'clear bit4, reg 137
      I2cwbyte 0
      I2cstop                                     'Unfreeze finished

      I2cstart                                    'Load DCO with new frequency
      I2cwbyte bytSi570WriteAdr
      I2cwbyte 135                                'set bit6, reg 135
      I2cwbyte &H40
      I2cstop
   End If
End Sub

'==============================================================================
Sub CalcSi570Addr
'Calculate read & write addresses from Si570 base address bytAddrSi570 (7 bit)
'Shift 7 bits 1 to left, write: bit0 = 0, read: bit0 = 1

   bytSi570WriteAdr = bytAddrSi570                'Si570 slave address
   Shift bytSi570WriteAdr , Left , 1              'Make 8 bit byte
   bytSi570WriteAdr.0 = 0                         'Si570 I2C write mode address
   bytSi570ReadAdr = bytSi570WriteAdr
   bytSi570ReadAdr.0 = 1                          'Si570 I2C read mode address
End Sub

'==============================================================================
Sub SetSmoothLimits
'Set the +/- 3,500 ppm limits of LO frequency for smooth tuning
'Within these limits only the DCO needs to reprogrammed
'so the output clock will have no interruption (smooth tuning)
'Input:  lngLo, current LO frequency
'Output: lngCenterFreq:          Center frequency configuration
'        lngLimitHi, lngLimitLo: upper/lower smooth tuning limits

   lngCenterFreq = lngLO                          'Current LO freq.
   sngTmp0 = lngLO
   sngTmp0 = sngTmp0 * 0.0035                     '3,500 ppm
   lngTmp0 = sngTmp0
   lngLimitHi = lngCenterFreq + lngTmp0           'Upper limit smooth tuning
   lngLimitLo = lngCenterFreq - lngTmp0           'Upper limit smooth tuning
   bitSmooth = 0                                  'No smooth tuning
End Sub

'==============================================================================
Sub CheckSmoothLimits
'Check the +/- 3,500 ppm limits of LO frequency for smooth tuning
'Within these limits only the DCO needs to reprogrammed
'so the output clock will have no interruption (smooth tuning)
'Input:  lngLo, current LO frequency
'Output: bitSmooth = 1: lngLO is within smooth tuning limits
'                  = 0: else, full LO freq. calculation is needed

   If lngLO >= lngLimitLo And lngLO <= lngLimitHi Then
      bitSmooth = 1                               'Smooth tuning
   Else
      Call SetSmoothLimits                        'set new limits
   End If
End Sub

'==============================================================================
Sub EEPROMInit
'Initialize EEPROM if empty
'Sto key pressed on power up: initialize anyway


   If Pin_KeySto = 1 Then                         'Sto key not pressed
      ReadEEPROM bytTmp0 , 1                      'If empty, value read=255
      If bytTmp0 = 1 Then Exit Sub                'Is already initialized
   End If

   CLS
   Locate 1 , 1
   LCD "Init EEPROM"
   Locate 2 , 1
   Lcd "Please wait"


   'VFO data
   wrdAddr = 7                                    'EEPROM address 7
   wrdSave = lngStep                              'Default tuning step (Hz)
   Call EEPROMWriteWrd(wrdSave)

   Incr wrdAddr                                   'EEPROM address 9
   WriteEEPROM bytLODiv , wrdAddr                 'Default mixer divider factor

   Incr wrdAddr                                   'EEPROM address 10
   WriteEEPROM bytAddrSi570 , wrdAddr             'Si570 slave address (hex55)

   Incr wrdAddr                                   'EEPROM address 11
   Call EEPROMWriteLng(lngFxtal)                  'used Si570 internal xtal freq (Hz)

   Incr wrdAddr                                   'EEPROM address 15
   WriteEEPROM bytI2C_BPF1 , wrdAddr              'I2C address 1st PCF8574A (BPF1)

   Incr wrdAddr                                   'EEPROM address 16
   WriteEEPROM bytI2C_BPF2 , wrdAddr              'I2C address 2nd PCF8574A (BPF2)

   'IF filter data
   wrdAddr = 20                                   'EEPROM address 20
   Call EEPROMWriteLng(lngIFSSB)                  'Default SSB IF center freq.(Hz)

   Incr wrdAddr                                   'EEPROM address 24
   Call EEPROMWriteWrd(wrdLSBOff)                 'Default LSB offset to SSB IF center (Hz)

   Incr wrdAddr                                   'EEPROM address 26
   WriteEEPROM bytLSBSign , wrdAddr               'Sign of LSB offset (-)

   Incr wrdAddr                                   'EEPROM address 27
   Call EEPROMWriteWrd(wrdUSBOff)                 'Default USB offset to SSB IF center (Hz)

   Incr wrdAddr                                   'EEPROM address 29
   WriteEEPROM bytUSBSign , wrdAddr               'Sign of LSB offset (+)

   Incr wrdAddr                                   'EEPROM address 30
   Call EEPROMWriteLng(lngIFCW)                   'Default CW IF center freq.(Hz)

   Incr wrdAddr                                   'EEPROM address 34
   Call EEPROMWriteWrd(wrdCWOff)                  'Default CW offset to CW IF center (Hz)

   Incr wrdAddr                                   'EEPROM address 36
   WriteEEPROM bytCWSign , wrdAddr                'Sign of CW offset (+)

   Incr wrdAddr                                   'EEPROM address 37
   Call EEPROMWriteLng(lngIFFM)                   'Default FM IF center freq.(Hz)

   'Band data
   wrdAddr = 69                                   'Band data @ EEPROM address 70+...
   For bytTmp0 = 1 To bytBandsMax                 'Loop the bands
         Incr wrdAddr
         bytTmp1 = bytBandUsed(bytTmp0)           'Band used: =1, else: =0
         WriteEEPROM bytTmp1 , wrdAddr

         Incr wrdAddr
         bytTmp1 = bytBand(bytTmp0)               'Band wave length (m)
         WriteEEPROM bytTmp1 , wrdAddr

         Incr wrdAddr
         bytTmp1 = bytBandBPF(bytTmp0)            'BPF number for this band
         WriteEEPROM bytTmp1 , wrdAddr

         Incr wrdAddr
         lngSave = lngBandLo(bytTmp0)             'Lower band limit (Hz)
         Call EEPROMWriteLng(lngSave)

         Incr wrdAddr
         lngSave = lngBandHi(bytTmp0)             'Upper band limit (Hz)
         Call EEPROMWriteLng(lngSave)
   Next bytTmp0

   bytTmp0 = 1
   WriteEEPROM bytTmp0 , 1                        'Done, write 1 to address 1
   Locate 2 , 1
   Lcd "Write done      "
   Waitms 500
End Sub

'==============================================================================
Sub EEPROMReadSettings
'Read settings from EEPROM (addresses 10 to 39)
'Input:  wrdAddr     first EEPROM address
'Output: lngStep     Freq. tuning step, all freq. in Hz
'        bytLODiv    mixer divider factor (1, 2, 4, 8)
'        bytAddrSi570 Si570 factory setting slave address
'        dblFxtal    Si570 internal xtal freq (MHz)
'        lngIFSSB    SSB IF center freq
'        lngIFCW     CW IF center freq
'        lngIFFM     FM IF center freq
'        wrdLSBOff   LSB offset relative to lngIFSSB
'        wrdUSBOff   LSB offset relative to lngIFSSB
'        wrdCWOff    CW offset relative to lngIFCW
'        bytLSBSign  1=Subtract Offset, 2=add
'        bytUSBSign  1=Subtract Offset, 2=add
'        bytCWSign   1=Subtract Offset, 2=add

   CLS
   Locate 1 , 1
   LCD "Read EEPROM     "

   wrdAddr = 7                                    'EEPROM start address 7
   Call EEPROMReadWrd(wrdSave)                    'Tuning step
   lngStep = wrdSave

   Incr wrdAddr                                   'EEPROM address 9
   ReadEEPROM bytTmp0 , wrdAddr                   'mixer divider factor
   bytLODiv = bytTmp0

   Incr wrdAddr                                   'EEPROM address 10
   ReadEEPROM bytTmp0 , wrdAddr                   'Si570 slave address
   bytAddrSi570 = bytTmp0

   Incr wrdAddr                                   'EEPROM address 11                                   '
   Call EEPROMReadLng(lngFxtal)                   'used Si570 internal xtal freq (Hz)
   dblFxtal = lngFxtal                            'Hz
   dblFxtal = dblFxtal * 0.000001                 'MHz

   Incr wrdAddr                                   'EEPROM address 15
   ReadEEPROM bytTmp0 , wrdAddr                   'I2C address 1st PC8574A (BPF1)
   bytI2C_BPF1 = bytTmp0
   Incr wrdAddr                                   'EEPROM address 16
   ReadEEPROM bytTmp0 , wrdAddr                   'I2C address 2nd PC8574A (BPF2)
   bytI2C_BPF2 = BytTmp0

   wrdAddr = 20                                   'EEPROM address 20
   Call EEPROMReadLng(lngSave)                    'SSB IF center freq.
   lngIFSSB = lngSave

   Incr wrdAddr                                   'EEPROM address 24
   Call EEPROMReadWrd(wrdSave)                    'LSB offset from SSB IF center (Hz)
   wrdLSBOff = wrdSave

   Incr wrdAddr                                   'EEPROM address 26
   ReadEEPROM bytTmp0 , wrdAddr                   'Sign of LSB offset
   bytLSBSign = bytTmp0

   Incr wrdAddr                                   'EEPROM address 27
   Call EEPROMReadWrd(wrdSave)                    'USB offset from SSB IF center (Hz)
   wrdUSBOff = wrdSave

   Incr wrdAddr                                   'EEPROM address 29
   ReadEEPROM bytTmp0 , wrdAddr                   'Sign of LSB offset
   bytUSBSign = bytTmp0

   Incr wrdAddr                                   'EEPROM address 30
   Call EEPROMReadLng(lngSave)                    'CW IF center freq.
   lngIFCW = lngSave

   Incr wrdAddr                                   'EEPROM address 34
   Call EEPROMReadWrd(wrdSave)                    'CW offset from CW IF center (Hz)
   wrdCWOff = wrdSave

   Incr wrdAddr                                   'EEPROM address 36
   ReadEEPROM bytTmp0 , wrdAddr                   'Sign of CW offset
   bytCWSign = bytTmp0

   Incr wrdAddr                                   'EEPROM address 37
   Call EEPROMReadLng(lngSave)                    'CW IF center freq.
   lngIFFM = lngSave
End Sub

'==============================================================================
Sub EEPROMReadBands(byVal bytAll As Byte)
'Read band data from EEPROM (from bands start address 70)
'Input:  bytAll      =1: all bands, =0: only used bands
'Output: bytBands    Number of used bands
'                    band index = 1, 2,...bytBands
'        bytBand     band wave length (m)
'        bytBandBPF  Number of bandpass filter for the band
'                    BPF index = 1, 2,...bytBPFs
'        lngBandLo   Lower band limits (Hz)
'        lngBandHi   Higher band limits (Hz)


   Locate 2 , 1
   Lcd "Bands: "
   bytBPFs = 0
   bytBands = 0
   wrdAddr = 69                                   'Band data @ EEPROM address 70+...
   For bytTmp0 = 1 To bytBandsMax                 'Loop the bands
       Incr wrdAddr
       ReadEEPROM bytTmp1 , wrdAddr               'Band used: =1, else: =0
       If bytAll = 0 And bytTmp1 = 0 Then         'Not used
            wrdAddr = wrdAddr + 10                'forget this band
       Else                                       'Read this band
            bytBands = bytBands + 1               'Band index
            Incr wrdAddr
            ReadEEPROM bytTmp1 , wrdAddr          'Band wave length (m)
            bytBand(bytBands) = bytTmp1

            Incr wrdAddr
            ReadEEPROM bytTmp1 , wrdAddr          'BPF number
            bytBandBPF(bytBands) = bytTmp1
            'Count used bandpass filters, may be less then bytBands
            If bytTmp1 > bytBPFs Then
               bytBPFs = bytTmp1                  'max BPF number
            End If

            Incr wrdAddr
            Call EEPROMReadLng(lngTmp0)           'Lower band limit (Hz)
            lngBandLo(bytBands) = lngTmp0

            Incr wrdAddr
            Call EEPROMReadLng(lngTmp0)           'Upper band limit (Hz)
            lngBandHi(bytBands) = lngTmp0
       End If
   Next bytTmp0
   If bytBands > 0 Then
      Locate 2 , 8
      LCD bytBands ; " found"
      Wait 1
   Else
      Locate 2 , 1
      LCD "No bands found!"
      Wait 5
   End If
   CLS
End Sub

'==============================================================================
Sub EEPROMWriteLng(lngLong As Long)
'Decode lngLong to 4 bytes for writing to EEPROM
'Input:  lngLong to be decoded & written to EEPROM
'        wrdAddr first EEPROM address
'        bytPart(i), 1...4, decoded bytes to be written to EEPROM
'        Most significant byte: bytPart(4)

   wrdTmp0 = HighW(lngLong)                       'High word of lngLong
   bytPart(4) = High(wrdTmp0)                     'High byte of wrdTmp0
   bytPart(3) = Low(wrdTmp0)                      'Low byte of wrdTmp0
   wrdTmp0 = lngLong                              'low word of lngLong
   bytPart(2) = High(wrdTmp0)                     'as above
   bytPart(1) = Low(wrdTmp0)
   WriteEEPROM bytPart(4) , wrdAddr               'write most high byte
   Incr wrdAddr                                   'Address 2. Byte
   WriteEEPROM bytPart(3) , wrdAddr
   Incr wrdAddr                                   'Address 3. Byte
   WriteEEPROM bytPart(2) , wrdAddr
   Incr wrdAddr                                   'Address 4. Byte
   WriteEEPROM bytPart(1) , wrdAddr
End Sub

'==============================================================================
Sub EEPROMReadLng(lngLong As Long)
'Encode 4 bytes read from EEPROM to lngLong
'Input:  wrdAddr first EEPROM address
'Output: lngLong encoded long variable
'        bytPart(i), 1...4, decoded bytes read from EEPROM
'        Most significant byte: bytPart(4)

   ReadEEPROM bytPart(4) , wrdAddr
   Incr wrdAddr
   ReadEEPROM bytPart(3) , wrdAddr
   Incr wrdAddr
   ReadEEPROM bytPart(2) , wrdAddr
   Incr wrdAddr
   ReadEEPROM bytPart(1) , wrdAddr
   wrdTmp0 = bytPart(4)                           'High byte of wrdTmp0
   Shift wrdTmp0 , Left , 8                       'Shift to most significant byte
   wrdTmp0 = wrdTmp0 + bytPart(3)                 'Add low byte
   lngLong = wrdTmp0                              'High word of lngLong
   Shift lngLong , Left , 16                      'Shift to most significant byte
   wrdTmp0 = bytPart(2)                           'High byte of wrdTmp0
   Shift wrdTmp0 , Left , 8                       'Shift to most significant byte
   wrdTmp0 = wrdTmp0 + bytPart(1)                 'Add low byte
   lngLong = lngLong + wrdTmp0                    'Add low word
End Sub

'==============================================================================
Sub EEPROMWriteWrd(wrdWord As Word)
'Decode wrdWord to 2 bytes for writing to EEPROM
'Input:  wrdWord to be decoded & written to EEPROM
'        wrdAddr first EEPROM address
'Output: bytPart(i), 1...2, decoded bytes to be written to EEPROM
'        Most significant byte: bytPart(2)

   bytPart(2) = High(wrdWord)                     'High byte of wrdWord
   bytPart(1) = Low(wrdWord)                      'Low byte of wrdWord
   WriteEEPROM bytPart(2) , wrdAddr               'write high byte
   Incr wrdAddr                                   'Address 2. Byte
   WriteEEPROM bytPart(1) , wrdAddr
End Sub

'==============================================================================
Sub EEPROMReadWrd(wrdWord As Word)
'Encode 2 bytes read from EEPROM to wrdWord
'Input:  wrdAddr first EEPROM address
'Output: wrdWord encoded unsigned integer (word) variable
'        bytPart(i), 1...2, decoded bytes read from EEPROM
'        Most significant byte: bytPart(2)

   ReadEEPROM bytPart(2) , wrdAddr
   Incr wrdAddr
   ReadEEPROM bytPart(1) , wrdAddr
   wrdWord = bytPart(2)                           'High byte of wrdWord
   Shift wrdWord , Left , 8                       'Shift to most significant byte
   wrdWord = wrdWord + bytPart(1)                 'Add low byte
End Sub